#!/usr/bin/perl
use strict;
use Data::Dumper;
use AnyEvent;
use EV;
use AnyEvent::Filesys::Notify;
use Fcntl qw(:flock SEEK_END);
use IO::Handle;
use DBI;
use JSON;
use Log::Syslog::Fast qw(:all);
use Config::JSON;
use Log::Log4perl;
use Getopt::Std qw(getopts);
use Net::Server::Daemonize qw(daemonize);

my %Opts;
getopts('c:D', \%Opts);

my $config_file = 'broweb.conf';
if ($Opts{c}){
	die('Invalid config file: ' . $Opts{c}) unless -f $Opts{c};
	$config_file = $Opts{c};
}
my $Conf = new Config::JSON($config_file) or die($!);

my $debug_level = $Conf->get('debug_level') ? $Conf->get('debug_level') : 'INFO';
	
# Setup logger
my $log_conf;
if ($Conf->get('web_logdir')){
	my $log_file = $Conf->get('web_logdir') . '/broweb.log';
	$log_conf = qq(
		log4perl.category.App       = $debug_level, File
		log4perl.appender.File			 = Log::Log4perl::Appender::File
		log4perl.appender.File.filename  = $log_file
		log4perl.appender.File.syswrite = 1
		log4perl.appender.File.recreate = 1
		log4perl.appender.File.layout = Log::Log4perl::Layout::PatternLayout
		log4perl.appender.File.layout.ConversionPattern = * %p [%d] %F (%L) %M %P %m%n
		log4perl.filter.ScreenLevel               = Log::Log4perl::Filter::LevelRange
  		log4perl.filter.ScreenLevel.LevelMin  = INFO
  		log4perl.filter.ScreenLevel.LevelMax  = ERROR
  		log4perl.filter.ScreenLevel.AcceptOnMatch = true
  		log4perl.appender.Screen         = Log::Log4perl::Appender::Screen
		log4perl.appender.Screen.Filter = ScreenLevel 
		log4perl.appender.Screen.stderr  = 1
		log4perl.appender.Screen.layout = Log::Log4perl::Layout::PatternLayout
		log4perl.appender.Screen.layout.ConversionPattern = * %p [%d] %F (%L) %M %P %m%n
	);
}
else {
	$log_conf = qq(
		log4perl.category.App       = $debug_level, Screen
		log4perl.filter.ScreenLevel               = Log::Log4perl::Filter::LevelRange
  		log4perl.filter.ScreenLevel.LevelMin  = TRACE
  		log4perl.filter.ScreenLevel.LevelMax  = ERROR
  		log4perl.filter.ScreenLevel.AcceptOnMatch = true
  		log4perl.appender.Screen         = Log::Log4perl::Appender::Screen
		log4perl.appender.Screen.Filter = ScreenLevel 
		log4perl.appender.Screen.stderr  = 1
		log4perl.appender.Screen.layout = Log::Log4perl::Layout::PatternLayout
		log4perl.appender.Screen.layout.ConversionPattern = * %p [%d] %F (%L) %M %P %m%n
	);
}

Log::Log4perl::init( \$log_conf ) or die("Unable to init logger\n");
my $Log = Log::Log4perl::get_logger('App')
	  or die("Unable to init logger\n");

my $Fields = {}; 
my $Filehandles = {};
my $Stats = {};
my $Start_time = time();

opendir(DIR, $Conf->get('bro_log_dir'));
while( my $file = readdir(DIR)){
	if ($file =~ /\.log$/){
		my $filename = $Conf->get('bro_log_dir') . '/' . $file;
		$Filehandles->{$filename} = new IO::Handle();
		my $fh;
		open($fh, $filename);
		my $first_line = <$fh>;
		chomp($first_line);
		$first_line =~ s/^#\s+//;
		my @fields = split(/\t/, $first_line);
		next unless @fields;
		$Fields->{$filename} = [ @fields ]; 
		$Filehandles->{$filename}->fdopen($fh, 'r');
		seek($Filehandles->{$filename}, 0, SEEK_END);
	}
}
close(DIR);

$SIG{TERM} = $SIG{INT} = \&print_stats;

sub print_stats {
	my $runtime = time() - $Start_time;
	foreach my $file (keys %$Stats){
		print $file . ' ' . $Stats->{$file} / $runtime . ' events/sec' . "\n";
	}
	exit;
}

if ($Opts{D}){
	print "Daemonizing...\n";
	daemonize('root', 'root', '/var/run/broweb.pid');
}

my $Dbh;
if ($Conf->get('db')){
	$Dbh = DBI->connect($Conf->get('db/dsn'), $Conf->get('db/username'), $Conf->get('db/password')) or die($DBI::errstr);
}
my $Syslog;
if ($Conf->get('syslog')){
	my $conf = $Conf->get('syslog');
	no strict qw(refs);
	$Syslog = Log::Syslog::Fast->new($conf->{proto}->(), $conf->{host}, $conf->{port}, $conf->{facility}->(),
		$conf->{level}->(), $conf->{sender}, $conf->{program});
} 
	

my $notifier = AnyEvent::Filesys::Notify->new(
	dirs => [ $Conf->get('bro_log_dir') ],
	filter => sub { shift =~ /\.log$/ },
	cb     => sub {
		my (@events) = @_;
		foreach my $event (@events){
			if ($event->is_created){
				$Log->info('Reading from new file ' . $event->path);
				$Filehandles->{$event->path} = new IO::Handle();
				my $fh;
				open($fh, $event->path);
				$Filehandles->{$event->path}->fdopen($fh, 'r');
			}
			unless ($event->is_deleted){
				my $data = _read_events($Filehandles->{$event->path}, $event->path);
				if ($Conf->get('db')){
					_write_events_db($data->{events});
				}
				if ($Conf->get('syslog')){
					_write_events_syslog($data->{lines});
				}
			}
		}
	},
);
	
EV::loop;

sub _read_events {
	my $fh = shift;
	my $filename = shift;
	return unless $Fields->{$filename};
	my @fields = @{ $Fields->{$filename} };
	
	$filename =~ /\/(\w+)\.log$/;
	my $type = $1;
	
	my @lines = $fh->getlines();
	my @hashes;
	if ($Conf->get('db')){
		foreach my $line (@lines){
			chomp($line);
			my @values = split(/\t/, $line);
			$Stats->{$filename}++;
			my $hash = { type => $type };
			for (my $i = 0; $i < $#fields; $i++){
				next unless defined $values[$i] and $values[$i] ne '-';
				$hash->{ $fields[$i] } = $values[$i];
			}
			push @hashes, $hash;
		}
	}
	@lines = map { "$type\t" . $_ } @lines;
	return { lines => \@lines, events => \@hashes };
}

sub _write_events_db {
	my $events = shift;
	
	my ($query, $sth);
	foreach my $event (@$events){
		$query = 'INSERT INTO data (data) VALUES (?)';
		$sth = $Dbh->prepare($query);
		$sth->execute(JSON::encode_json($event));
		
		unless ($sth->rows){
			return 0;
		}
		
		my $id = $Dbh->last_insert_id(undef, undef, 'data', 'data');
		
		$query = 'INSERT IGNORE INTO kv (data_id, k, v) VALUES ';
		my @placeholders;
		foreach my $key (keys %$event){
			$query .= '(?,?,?),';
			push @placeholders, $id, $key, $event->{$key};
		}
		chop($query);
		$sth = $Dbh->prepare($query);
		$sth->execute(@placeholders);
	}
}

sub _write_events_syslog {
	my $lines = shift;
	foreach my $line (@$lines){
		chomp($line);
		my ($program, @terms) = split(/\t/, $line);
		$program = 'bro_' . $program;
		$Syslog->set_name($program);
		$Syslog->send(join('|', @terms));
		#print $program . ' ' . join('|', @terms) . "\n";
	}
}

__END__
CREATE TABLE data (
	id INT UNSIGNED NOT NULL PRIMARY KEY AUTO_INCREMENT,
	data TEXT
);
CREATE TABLE kv (
	data_id INT UNSIGNED,
	k VARCHAR(255) NOT NULL,
	v TEXT,
	KEY (data_id),
	KEY (k),
	KEY (v(8))
);